<?php declare(strict_types=1);
/*
 * This file is part of PHPUnit.
 *
 * (c) Sebastian Bergmann <sebastian@phpunit.de>
 *
 * For the full copyright and license information, please view the LICENSE
 * file that was distributed with this source code.
 */
namespace PHPUnit\TestFixture;

class SingletonClass
{
    public static function getInstance(): void
    {
    }

    protected function __construct()
    {
    }

    public function __sleep(): array
    {
        return [];
    }

    public function __wakeup(): void
    {
    }

    private function __clone()
    {
    }

    public function doSomething(): void
    {
    }
}
